package ru.dbotthepony.kstarbound.client.gl

import ru.dbotthepony.kstarbound.client.gl.shader.GLShader
import ru.dbotthepony.kstarbound.client.gl.shader.GLShaderProgram
import ru.dbotthepony.kstarbound.client.gl.shader.GLTransformableColorableProgram
import kotlin.properties.ReadOnlyProperty
import kotlin.reflect.KProperty

private class SimpleProgram<T : GLShaderProgram>(private val name: String, private val factory: (GLStateTracker, Array<out GLShader>) -> T) : ReadOnlyProperty<GLPrograms, T> {
	private var value: T? = null

	override fun getValue(thisRef: GLPrograms, property: KProperty<*>): T {
		val value = value

		if (value != null) {
			return value
		}

		val vertex = GLShader.internalVertex("shaders/$name.vsh")
		val fragment = GLShader.internalFragment("shaders/$name.fsh")

		val newValue = factory.invoke(thisRef.state, arrayOf(vertex, fragment))

		this.value = newValue

		vertex.unlink()
		fragment.unlink()

		if (!newValue.linked) {
			newValue.link()
		}

		return newValue
	}
}

class GLPrograms(val state: GLStateTracker) {
	val tile by SimpleProgram("tile", ::GLTransformableColorableProgram)
	val font by SimpleProgram("font", ::GLTransformableColorableProgram)
	val flat by SimpleProgram("flat", ::GLTransformableColorableProgram)
}
