package ru.dbotthepony.kstarbound.client.gl.shader

import org.lwjgl.opengl.GL41.*
import ru.dbotthepony.kstarbound.client.gl.checkForGLError
import ru.dbotthepony.kvector.api.IFloatMatrix
import ru.dbotthepony.kvector.api.IStruct3f
import ru.dbotthepony.kvector.api.IStruct4f
import java.nio.ByteBuffer
import java.nio.ByteOrder

class GLUniformLocation(val program: GLShaderProgram, val name: String, val pointer: Int) {
	fun set(value: IStruct4f): GLUniformLocation {
		program.state.ensureSameThread()
		val (v0, v1, v2, v3) = value
		glProgramUniform4f(program.pointer, pointer, v0, v1, v2, v3)
		return this
	}

	fun set(value: IStruct3f): GLUniformLocation {
		program.state.ensureSameThread()
		val (v0, v1, v2) = value
		glProgramUniform3f(program.pointer, pointer, v0, v1, v2)
		return this
	}

	fun set(value: Int): GLUniformLocation {
		program.state.ensureSameThread()
		glProgramUniform1i(program.pointer, pointer, value)
		return this
	}

	private val buff3x3 by lazy { ByteBuffer.allocateDirect(4 * 3 * 3).also { it.order(ByteOrder.nativeOrder()) }.asFloatBuffer() }
	private val buff4x4 by lazy { ByteBuffer.allocateDirect(4 * 4 * 4).also { it.order(ByteOrder.nativeOrder()) }.asFloatBuffer() }

	fun set(value: IFloatMatrix<*>): GLUniformLocation {
		program.state.ensureSameThread()

		if (value.rows == 3 && value.columns == 3) {
			// Матрица 3x3
			buff3x3.position(0)
			value.write(buff3x3)
			buff3x3.position(0)
			glProgramUniformMatrix3fv(program.pointer, pointer, false, buff3x3)
			checkForGLError()
		} else if (value.rows == 4 && value.columns == 4) {
			// Матрица 4x4
			buff4x4.position(0)
			value.write(buff4x4)
			buff4x4.position(0)
			glProgramUniformMatrix4fv(program.pointer, pointer, false, buff4x4)
			checkForGLError()
		} else {
			throw IllegalArgumentException("Can not use matrix with these dimensions: ${value.columns}x${value.rows}")
		}

		return this
	}
}