package ru.dbotthepony.kstarbound.client.render

import org.lwjgl.opengl.GL46.*
import ru.dbotthepony.kstarbound.PIXELS_IN_STARBOUND_UNITf
import ru.dbotthepony.kstarbound.client.ClientChunk
import ru.dbotthepony.kstarbound.client.gl.GLStateTracker
import ru.dbotthepony.kstarbound.client.gl.vertex.QuadTransformers
import ru.dbotthepony.kstarbound.world.entities.Entity
import ru.dbotthepony.kstarbound.world.entities.projectile.Projectile
import ru.dbotthepony.kvector.matrix.Matrix4fStack
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import java.io.Closeable

/**
 * Pseudo Z position for entities, for them to appear behind tile geometry,
 * but in front of background walls geometry
 */
const val Z_LEVEL_ENTITIES = 30000

/**
 * Базовый класс, отвечающий за отрисовку определённого ентити в мире
 *
 * Считается, что процесс отрисовки ограничен лишь одним слоем (т.е. отрисовка происходит в один проход)
 */
open class EntityRenderer(val state: GLStateTracker, val entity: Entity, open var chunk: ClientChunk?) : Closeable {
	open val renderPos: Vector2d get() = entity.position

	open fun render(stack: Matrix4fStack) {

	}

	open fun renderDebug() {
		if (chunk?.world?.client?.settings?.debugCollisions == true) {
			state.quadWireframe(entity.movement.worldAABB)
		}
	}

	open val layer: Int = Z_LEVEL_ENTITIES

	override fun close() {

	}

	companion object {
		fun getRender(state: GLStateTracker, entity: Entity, chunk: ClientChunk? = null): EntityRenderer {
			return when (entity) {
				is Projectile -> ProjectileRenderer(state, entity, chunk)
				else -> EntityRenderer(state, entity, chunk)
			}
		}
	}
}

open class ProjectileRenderer(state: GLStateTracker, entity: Projectile, chunk: ClientChunk?) : EntityRenderer(state, entity, chunk) {
	private val def = entity.def
	private val texture = state.loadNamedTextureSafe(def.image.texture)
	private val animator = FrameSetAnimator(def.image, def.animationCycle, entity.def.animationLoops)

	init {
		texture.textureMagFilter = GL_NEAREST
	}

	override fun render(stack: Matrix4fStack) {
		state.shaderVertexTexture.use()
		state.shaderVertexTexture.transform.set(stack.last)
		state.activeTexture = 0
		state.shaderVertexTexture["_texture"] = 0
		texture.bind()

		animator.advance()

		val builder = state.flat2DTexturedQuads.small

		builder.begin()

		val (u0, v0) = texture.pixelToUV(animator.frameObj.texturePosition)
		val (u1, v1) = texture.pixelToUV(animator.frameObj.textureEndPosition)

		val width = (animator.frameObj.width / PIXELS_IN_STARBOUND_UNITf) / 2f
		val height = (animator.frameObj.height / PIXELS_IN_STARBOUND_UNITf) / 2f

		builder.quadRotatedZ(-width, -height, width, height, 5f, 0f, 0f, entity.movement.angle, QuadTransformers.uv(u0, v0, u1, v1))

		builder.upload()
		builder.draw()
	}
}
