package ru.dbotthepony.kstarbound.world.phys

import ru.dbotthepony.kstarbound.world.CHUNK_SHIFT
import ru.dbotthepony.kstarbound.world.CHUNK_SIZE_FF
import ru.dbotthepony.kstarbound.world.TileState
import ru.dbotthepony.kvector.vector.ndouble.Vector2d

private data class TileExposure(
	var pos: Vector2d,
	val left: Boolean,
	val right: Boolean,
	val up: Boolean,
	val down: Boolean,
)

private class TileFlooder(
	private val tiles: Array<TileState?>,
	private val seen: BooleanArray,
	rootx: Int,
	rooty: Int
) {
	/**
	 * Tile positions which have at least one face free of neighbours
	 */
	val exposed = ArrayList<TileExposure>()

	private fun get(x: Int, y: Int): Boolean {
		if (x < 0 || x > CHUNK_SIZE_FF) {
			return false
		}

		if (y < 0 || y > CHUNK_SIZE_FF) {
			return false
		}

		return tiles[x or (y shl CHUNK_SHIFT)] != null
	}

	private fun visit(x: Int, y: Int) {
		if (seen[x or (y shl CHUNK_SHIFT)])
			return

		seen[x or (y shl CHUNK_SHIFT)] = true

		val left = get(x - 1, y)
		val right = get(x + 1, y)
		val up = get(x, y + 1)
		val down = get(x, y - 1)

		if (!left || !right || !up || !down) {
			exposed.add(TileExposure(Vector2d(x.toDouble() + 0.5, y.toDouble() + 0.5), !left, !right, !up, !down))
		}

		if (left) {
			visit(x - 1, y)
		}

		if (right) {
			visit(x + 1, y)
		}

		if (up) {
			visit(x, y + 1)
		}

		if (down) {
			visit(x, y - 1)
		}
	}

	init {
		visit(rootx, rooty)
	}
}